from __future__ import annotations

import typing

from di.api.providers import CallableProvider, CoroutineProvider
from di.concurrency import as_async
from di.dependent import Dependent

from xpresso.dependencies._dependencies import Depends, DependsMarker

Endpoint = typing.Union[CallableProvider[typing.Any], CoroutineProvider[typing.Any]]


class EndpointDependent(Dependent[typing.Any]):
    def __init__(
        self,
        endpoint: Endpoint,
        sync_to_thread: bool = False,
    ) -> None:
        if sync_to_thread:
            endpoint = as_async(endpoint)
        super().__init__(
            call=endpoint,
            scope="endpoint",
            use_cache=False,
            wire=True,
        )

    def get_default_marker(self) -> DependsMarker[None]:
        return Depends()



import contextlib
import inspect
import typing
from types import TracebackType

from di.api.dependencies import DependentBase
from di.api.providers import DependencyProvider
from di.container import Container
from di.dependent import Dependent

from xpresso._utils.typing import Annotated, get_args, get_origin


def get_type(param: inspect.Parameter) -> type:
    if get_origin(param.annotation) is Annotated:
        return next(iter(get_args(param.annotation)))
    return param.annotation


class DependencyOverrideManager:
    _stacks: typing.List[contextlib.ExitStack]

    def __init__(self, container: Container) -> None:
        self._container = container
        self._stacks = []

    def __setitem__(
        self, target: DependencyProvider, replacement: DependencyProvider
    ) -> None:
        def hook(
            param: typing.Optional[inspect.Parameter],
            dependent: DependentBase[typing.Any],
        ) -> typing.Optional[DependentBase[typing.Any]]:
            if not isinstance(dependent, Dependent):
                return None
            scope = dependent.scope
            dep = Dependent(
                replacement,
                scope=scope,  # type: ignore[arg-type]
                use_cache=dependent.use_cache,
                wire=dependent.wire,
            )
            if param is not None and param.annotation is not param.empty:
                type_ = get_type(param)
                if type_ is target:
                    return dep
            if dependent.call is not None and dependent.call is target:
                return dep
            return None

        cm = self._container.bind(hook)
        if self._stacks:
            self._stacks[-1].enter_context(cm)

    def __enter__(self) -> "DependencyOverrideManager":
        self._stacks.append(contextlib.ExitStack().__enter__())
        return self

    def __exit__(
        self,
        __exc_type: typing.Optional[typing.Type[BaseException]],
        __exc_value: typing.Optional[BaseException],
        __traceback: typing.Optional[TracebackType],
    ) -> typing.Optional[bool]:
        return self._stacks.pop().__exit__(__exc_type, __exc_value, __traceback)



from typing import Optional

from di.container._state import ContainerState
from starlette.responses import Response


class XpressoHTTPExtension:
    __slots__ = ("di_container_state", "response", "response_sent")

    di_container_state: ContainerState
    response: Optional[Response]
    response_sent: bool

    def __init__(self, di_state: ContainerState) -> None:
        self.di_container_state = di_state
        self.response = None
        self.response_sent = False


class XpressoWebSocketExtension:
    __slots__ = ("di_container_state",)

    di_container_state: ContainerState

    def __init__(self, di_state: ContainerState) -> None:
        self.di_container_state = di_state



import contextlib
import inspect
import typing

import xpresso.openapi.models as openapi_models
from xpresso._utils.pydantic_utils import model_field_from_param
from xpresso._utils.typing import Annotated, get_args, get_origin
from xpresso.binders._binders.utils import (
    Consumer,
    ConsumerContextManager,
    wrap_consumer_as_cm,
)
from xpresso.binders.api import ModelNameMap, SupportsExtractor, SupportsOpenAPI
from xpresso.binders.dependents import Binder, BinderMarker
from xpresso.exceptions import HTTPException, RequestValidationError
from xpresso.requests import HTTPConnection, Request

RequestConsumer = Consumer[Request]
RequestConsumerContextManger = ConsumerContextManager[Request]


def get_binders_from_union_annotation(param: inspect.Parameter) -> typing.List[Binder]:
    providers: typing.List[Binder] = []
    annotation = param.annotation
    origin = get_origin(annotation)
    assert origin is Annotated  # if we got called, this must be true
    annotation = next(iter(get_args(annotation)))
    origin = get_origin(annotation)
    if origin is not typing.Union:
        raise TypeError("Outermost type must be a Union")
    args = get_args(annotation)
    for arg in args:
        origin = get_origin(arg)
        assert origin is Annotated  # if we got called, this must be true
        marker = next((m for m in get_args(arg) if isinstance(m, BinderMarker)), None)
        if marker is None:
            raise TypeError(f"Type annotation is missing body marker: {arg}")
        providers.append(marker.register_parameter(param.replace(annotation=arg)))
    return providers


class BodyOpenAPI(typing.NamedTuple):
    description: typing.Optional[str]
    required: bool
    providers: typing.List[SupportsOpenAPI]

    def get_models(self) -> typing.List[type]:
        return [model for p in self.providers for model in p.get_models()]

    def modify_operation_schema(
        self,
        model_name_map: ModelNameMap,
        operation: openapi_models.Operation,
        components: openapi_models.Components,
    ) -> None:
        for provider in self.providers:
            provider.modify_operation_schema(model_name_map, operation, components)


class BodyOpenAPIMarker(typing.NamedTuple):
    description: typing.Optional[str]

    def register_parameter(self, param: inspect.Parameter) -> SupportsOpenAPI:
        field = model_field_from_param(param)
        return BodyOpenAPI(
            providers=[b.openapi for b in get_binders_from_union_annotation(param)],
            required=field.required is not False,
            description=self.description,
        )


SupportsExtractorCM = typing.Callable[
    [HTTPConnection], typing.AsyncContextManager[typing.Any]
]


class Extractor(typing.NamedTuple):
    extractors: typing.Iterable[SupportsExtractorCM]

    async def extract(
        self, connection: HTTPConnection
    ) -> typing.AsyncIterator[typing.Any]:
        assert isinstance(connection, Request)
        errors: "typing.List[typing.Union[HTTPException, RequestValidationError]]" = []
        for extractor in self.extractors:
            try:
                async with extractor(connection) as res:
                    yield res
                return
            except (HTTPException, RequestValidationError) as error:
                errors.append(error)
        # if any body accepted the request but didn't pass validation, return the error from that one
        for err in errors:
            if err.status_code == 422:
                raise err
        # otherwise, just raise the first error we found
        # this is somewhat arbitrary, but there really is no good way to "merge"
        # all of the errors without making it confusing to the client
        # and leaking implementation details
        raise next(iter(errors))

    def __hash__(self) -> int:
        return id(self)

    def __eq__(self, __o: object) -> bool:
        return False


class ExtractorMarker(typing.NamedTuple):
    def register_parameter(self, param: inspect.Parameter) -> SupportsExtractor:
        extractors: typing.List[SupportsExtractorCM] = []
        for binder in get_binders_from_union_annotation(param):
            extractor = binder.extractor.extract
            if inspect.isasyncgenfunction(extractor):
                extractors.append(
                    contextlib.asynccontextmanager(extractor)  # type: ignore[arg-type]
                )
            else:
                extractors.append(wrap_consumer_as_cm(extractor))
        return Extractor(extractors)



from typing import Any, Callable, Iterable, Type

from starlette.datastructures import UploadFile as StarletteUploadFile


class UploadFile(StarletteUploadFile):
    @classmethod
    def __get_validators__(cls: Type["UploadFile"]) -> Iterable[Callable[..., Any]]:
        # this is required so that UploadFile can be a Pydantic field
        return iter(())

    async def read(self, size: int = -1) -> bytes:
        # this is implemented just to fix the return type annotation
        # which is always bytes
        return await super().read(size)  # type: ignore



import contextlib
import typing

T = typing.TypeVar("T")

Consumer = typing.Callable[[T], typing.Any]
ConsumerContextManager = typing.Callable[[T], typing.AsyncContextManager[typing.Any]]


def wrap_consumer_as_cm(consumer: Consumer[T]) -> ConsumerContextManager[T]:
    @contextlib.asynccontextmanager
    async def consume(request: T) -> typing.AsyncIterator[typing.Any]:
        res = await consumer(request)
        yield res

    return consume



import typing

from pydantic.error_wrappers import ErrorWrapper
from pydantic.fields import ModelField
from starlette.requests import HTTPConnection

from xpresso.exceptions import RequestValidationError, WebSocketValidationError
from xpresso.typing import Some


def validate_param_field(
    field: ModelField,
    name: str,
    in_: str,
    values: typing.Optional[Some],
    connection: HTTPConnection,
) -> typing.Any:
    """Validate after parsing. Only used by the top-level body"""
    if values is None:
        if field.required is False:
            return field.get_default()
        else:
            err = [
                ErrorWrapper(
                    ValueError(f"Missing required {in_} parameter"),
                    loc=(in_, name),
                )
            ]
            if connection.scope["type"] == "websocket":
                raise WebSocketValidationError(err)
            raise RequestValidationError(err)
    val, errs = field.validate(values.value, {}, loc=(in_, name))
    if errs:
        if isinstance(errs, ErrorWrapper):
            errs = [errs]
        errs = typing.cast(typing.List[ErrorWrapper], errs)
        if connection.scope["type"] == "websocket":
            raise WebSocketValidationError(errs)
        raise RequestValidationError(errs)
    return val


def validate_body_field(
    values: typing.Optional[Some],
    *,
    field: ModelField,
    loc: typing.Tuple[typing.Union[str, int], ...],
) -> typing.Any:
    """Validate after extraction. Should only be used by the top-level body"""
    if values is None:
        if field.required is False:
            return field.get_default()
        else:
            raise RequestValidationError(
                [ErrorWrapper(ValueError("Missing required value"), loc=loc)]
            )
    val, err_or_errors = field.validate(values.value, {}, loc=loc)
    if err_or_errors:
        errors: typing.List[ErrorWrapper]
        if isinstance(err_or_errors, ErrorWrapper):
            errors = [err_or_errors]
        else:
            errors = typing.cast(
                typing.List[ErrorWrapper], err_or_errors
            )  # already a list
        raise RequestValidationError(errors)
    return val



import typing

import starlette.requests
import starlette.responses
import starlette.routing
import starlette.types
import starlette.websockets
from di.api.dependencies import DependentBase
from di.api.executor import SupportsAsyncExecutor
from di.api.solved import SolvedDependent
from di.container import Container
from di.dependent import JoinedDependent
from di.executors import AsyncExecutor, ConcurrentAsyncExecutor

from xpresso._utils.asgi import XpressoWebSocketExtension
from xpresso._utils.endpoint_dependent import Endpoint, EndpointDependent
from xpresso._utils.scope_resolver import endpoint_scope_resolver
from xpresso.dependencies._dependencies import BoundDependsMarker, Scopes


class _WebSocketRoute:
    __slots__ = ("container", "dependent", "executor")

    def __init__(
        self,
        dependent: SolvedDependent[typing.Any],
        executor: SupportsAsyncExecutor,
        container: Container,
    ) -> None:
        self.dependent = dependent
        self.executor = executor
        self.container = container

    async def __call__(
        self,
        scope: starlette.types.Scope,
        receive: starlette.types.Receive,
        send: starlette.types.Send,
    ) -> None:
        xpresso_scope: XpressoWebSocketExtension = scope["extensions"]["xpresso"]
        ws = starlette.websockets.WebSocket(scope=scope, receive=receive, send=send)
        values: typing.Dict[typing.Any, typing.Any] = {
            starlette.websockets.WebSocket: ws,
            starlette.requests.HTTPConnection: ws,
        }
        async with self.container.enter_scope(
            "connection",
            state=xpresso_scope.di_container_state,
        ) as conn_state:
            async with self.container.enter_scope(
                "endpoint", state=conn_state
            ) as endpoint_state:
                await self.container.execute_async(
                    self.dependent,
                    values=values,
                    executor=self.executor,
                    state=endpoint_state,
                )


class WebSocketRoute(starlette.routing.WebSocketRoute):
    path: str

    def __init__(
        self,
        path: str,
        endpoint: Endpoint,
        *,
        name: typing.Optional[str] = None,
        dependencies: typing.Optional[
            typing.Iterable[typing.Union[DependentBase[typing.Any], BoundDependsMarker]]
        ] = None,
        execute_dependencies_concurrently: bool = False,
    ) -> None:
        super().__init__(  # type: ignore
            path=path,
            endpoint=endpoint,
            name=name,  # type: ignore[arg-type]
        )
        self.endpoint = endpoint
        self.dependencies = tuple(
            dep if isinstance(dep, DependentBase) else dep.as_dependent()
            for dep in dependencies or ()
        )
        self.execute_dependencies_concurrently = execute_dependencies_concurrently

    def prepare(
        self,
        container: Container,
        dependencies: typing.Iterable[DependentBase[typing.Any]],
    ) -> SolvedDependent[typing.Any]:
        self.dependent = container.solve(
            JoinedDependent(
                EndpointDependent(self.endpoint),
                siblings=[*dependencies, *self.dependencies],
            ),
            scopes=Scopes,
            scope_resolver=endpoint_scope_resolver,
        )
        executor: SupportsAsyncExecutor
        if self.execute_dependencies_concurrently:
            executor = ConcurrentAsyncExecutor()
        else:
            executor = AsyncExecutor()
        self.app = _WebSocketRoute(
            dependent=self.dependent,
            executor=executor,
            container=container,
        )
        return self.dependent



from typing import Any

from di.dependent import Injectable
from pydantic import BaseSettings

from xpresso.dependencies._dependencies import Scope


class BaseConfig(Injectable, BaseSettings):
    def __init_subclass__(
        cls, scope: Scope = "app", use_cache: bool = True, **kwargs: Any
    ) -> None:
        # Pydantic BaseSettings models cannot be wired because of how the grab values
        # from env vars
        # But most of the time you just want to load the entire thing from the environment,
        # so that's what we do here
        def call() -> Any:
            return cls()

        return super().__init_subclass__(
            call=call, scope=scope, use_cache=use_cache, **kwargs
        )



import typing
from functools import partial

from di.api.dependencies import DependentBase
from di.api.executor import SupportsAsyncExecutor
from di.api.solved import SolvedDependent
from di.container import Container
from di.dependent import JoinedDependent
from di.executors import AsyncExecutor, ConcurrentAsyncExecutor
from starlette.datastructures import URLPath
from starlette.requests import HTTPConnection, Request
from starlette.responses import JSONResponse, Response
from starlette.routing import BaseRoute, NoMatchFound, get_name  # type: ignore
from starlette.types import ASGIApp, Receive, Scope, Send

import xpresso.openapi.models as openapi_models
from xpresso._utils.asgi import XpressoHTTPExtension
from xpresso._utils.endpoint_dependent import Endpoint, EndpointDependent
from xpresso._utils.scope_resolver import endpoint_scope_resolver
from xpresso.dependencies._dependencies import BoundDependsMarker, Scopes
from xpresso.encoders import Encoder, JsonableEncoder
from xpresso.responses import ResponseSpec, ResponseStatusCode, TypeUnset


class NotPreparedError(Exception):
    pass


class _OperationApp(typing.NamedTuple):
    dependent: SolvedDependent[typing.Any]
    container: Container
    executor: SupportsAsyncExecutor
    response_factory: typing.Callable[[typing.Any], Response]
    response_encoder: typing.Optional[Encoder]

    async def __call__(
        self,
        scope: Scope,
        receive: Receive,
        send: Send,
    ) -> None:
        xpresso_scope: "XpressoHTTPExtension" = scope["extensions"]["xpresso"]
        request = Request(scope=scope, receive=receive, send=send)
        values: "typing.Dict[typing.Any, typing.Any]" = {
            Request: request,
            HTTPConnection: request,
        }
        async with self.container.enter_scope(
            "connection",
            xpresso_scope.di_container_state,
        ) as connection_state:
            async with connection_state.enter_scope("endpoint") as endpoint_state:
                endpoint_return = await self.container.execute_async(
                    self.dependent,
                    values=values,
                    executor=self.executor,
                    state=endpoint_state,
                )
                if isinstance(endpoint_return, Response):
                    response = endpoint_return
                else:
                    if self.response_encoder:
                        endpoint_return = self.response_encoder(endpoint_return)
                    response = self.response_factory(endpoint_return)
                xpresso_scope.response = response
            await response(scope, receive, send)
            xpresso_scope.response_sent = True


async def _not_prepared_app(*args: typing.Any) -> None:
    raise NotPreparedError(
        "Operation.prepare() was never called on this Operation."
        " Perhaps you tried to use an Xpresso Operation outside of an Xpresso App?"
    )


class Operation(BaseRoute):
    def __init__(
        self,
        endpoint: Endpoint,
        *,
        # OpenAPI parameters
        include_in_schema: bool = True,
        tags: typing.Optional[typing.Sequence[str]] = None,
        summary: typing.Optional[str] = None,
        description: typing.Optional[str] = None,
        deprecated: typing.Optional[bool] = None,
        operation_id: typing.Optional[bool] = None,
        servers: typing.Optional[typing.Sequence[openapi_models.Server]] = None,
        external_docs: typing.Optional[openapi_models.ExternalDocumentation] = None,
        responses: typing.Optional[
            typing.Mapping[ResponseStatusCode, ResponseSpec]
        ] = None,
        # xpresso params
        name: typing.Optional[str] = None,
        dependencies: typing.Optional[
            typing.Iterable[typing.Union[DependentBase[typing.Any], BoundDependsMarker]]
        ] = None,
        execute_dependencies_concurrently: bool = False,
        response_factory: typing.Optional[
            typing.Callable[[typing.Any], Response]
        ] = None,
        response_encoder: typing.Optional[Encoder] = JsonableEncoder(),
        sync_to_thread: bool = False,
        # responses
        response_status_code: int = 200,
        response_media_type: str = "application/json",
        response_model: typing.Any = TypeUnset,
        response_description: typing.Optional[str] = None,
        response_examples: typing.Optional[
            typing.Mapping[str, typing.Union[openapi_models.Example, typing.Any]]
        ] = None,
        response_headers: typing.Optional[
            typing.Mapping[str, typing.Union[openapi_models.ResponseHeader, str]]
        ] = None,
    ) -> None:
        # These fields mirror Starlette's Route
        self.endpoint = endpoint
        self.include_in_schema = include_in_schema
        self.name: str = get_name(endpoint) if name is None else name  # type: ignore
        self.tags = tuple(tags or ())
        self.summary = summary
        self.description = description
        self.deprecated = deprecated
        self.operation_id = operation_id
        self.servers = tuple(servers or ())
        self.external_docs = external_docs
        self.responses = dict(responses or {})
        self.response_status_code = response_status_code
        self.response_media_type = response_media_type
        self.response_model = response_model
        self.response_description = response_description
        self.response_examples = response_examples
        self.response_headers = response_headers
        self.dependencies = tuple(
            dep if isinstance(dep, DependentBase) else dep.as_dependent()
            for dep in dependencies or ()
        )
        self._app: ASGIApp = _not_prepared_app
        self._execute_dependencies_concurrently = execute_dependencies_concurrently
        self._response_factory = response_factory or partial(
            JSONResponse,
            media_type=response_media_type,
            status_code=response_status_code,
        )
        self._response_encoder = response_encoder
        self._sync_to_thread = sync_to_thread

    async def handle(
        self,
        scope: Scope,
        receive: Receive,
        send: Send,
    ) -> None:
        return await self._app(scope, receive, send)

    def prepare(
        self,
        container: Container,
        dependencies: typing.Iterable[DependentBase[typing.Any]],
    ) -> SolvedDependent[typing.Any]:
        self.dependent = container.solve(
            JoinedDependent(
                EndpointDependent(self.endpoint, sync_to_thread=self._sync_to_thread),
                siblings=[*dependencies, *self.dependencies],
            ),
            scopes=Scopes,
            scope_resolver=endpoint_scope_resolver,
        )
        executor: SupportsAsyncExecutor
        if self._execute_dependencies_concurrently:
            executor = ConcurrentAsyncExecutor()
        else:
            executor = AsyncExecutor()
        self._app = _OperationApp(  # type: ignore[assignment]
            container=container,
            dependent=self.dependent,
            executor=executor,
            response_encoder=self._response_encoder,
            response_factory=self._response_factory,
        )
        return self.dependent

    def url_path_for(self, name: str, **path_params: str) -> URLPath:
        if path_params:
            raise NoMatchFound(name, path_params)
        if name != self.name:
            raise NoMatchFound(name, path_params)
        return URLPath("/")

    def __eq__(self, __o: object) -> bool:
        return isinstance(__o, type(self)) and self.endpoint is __o.endpoint



import typing

import starlette.routing
import starlette.types
from di.api.dependencies import DependentBase
from di.api.providers import DependencyProvider
from di.api.providers import DependencyProvider as Endpoint

import xpresso.binders.dependents as dependents
import xpresso.openapi.models as openapi_models
from xpresso.dependencies._dependencies import DependsMarker
from xpresso.responses import ResponseSpec, ResponseStatusCode
from xpresso.routing.operation import Operation


class _PathApp(typing.NamedTuple):
    operations: typing.Mapping[str, Operation]

    def __call__(
        self,
        scope: starlette.types.Scope,
        receive: starlette.types.Receive,
        send: starlette.types.Send,
    ) -> typing.Awaitable[None]:
        return self.operations[scope["method"]].handle(scope, receive, send)


class Path(starlette.routing.Route):
    include_in_schema: bool

    def __init__(
        self,
        path: str,
        *,
        get: typing.Optional[typing.Union[Operation, Endpoint]] = None,
        head: typing.Optional[typing.Union[Operation, Endpoint]] = None,
        post: typing.Optional[typing.Union[Operation, Endpoint]] = None,
        put: typing.Optional[typing.Union[Operation, Endpoint]] = None,
        patch: typing.Optional[typing.Union[Operation, Endpoint]] = None,
        delete: typing.Optional[typing.Union[Operation, Endpoint]] = None,
        connect: typing.Optional[typing.Union[Operation, Endpoint]] = None,
        options: typing.Optional[typing.Union[Operation, Endpoint]] = None,
        trace: typing.Optional[typing.Union[Operation, Endpoint]] = None,
        redirect_slashes: bool = True,
        dependencies: typing.Optional[
            typing.Iterable[
                typing.Union[
                    DependentBase[typing.Any], DependsMarker[DependencyProvider]
                ]
            ]
        ] = None,
        # OpenAPI metadata
        include_in_schema: bool = True,
        name: typing.Optional[str] = None,
        summary: typing.Optional[str] = None,
        description: typing.Optional[str] = None,
        servers: typing.Optional[typing.Sequence[openapi_models.Server]] = None,
        parameters: typing.Optional[typing.Sequence[dependents.BinderMarker]] = None,
        responses: typing.Optional[
            typing.Mapping[ResponseStatusCode, ResponseSpec]
        ] = None,
        tags: typing.Optional[typing.Iterable[str]] = None,
    ) -> None:
        if not path.startswith("/"):
            raise ValueError("Routed paths must start with '/'")
        self.path = path
        self.redirect_slashes = redirect_slashes
        self.dependencies = tuple(
            dep if isinstance(dep, DependentBase) else dep.as_dependent()
            for dep in dependencies or ()
        )
        self.summary = summary
        self.description = description
        self.servers = tuple(servers or ())
        self.parameters = list(parameters or ())
        self.tags = tuple(tags or ())
        self.responses = dict(responses or {})

        operations: typing.Dict[str, Operation] = {}
        for operation_or_endpoint, method in (
            (get, "GET"),
            (head, "HEAD"),
            (post, "POST"),
            (put, "PUT"),
            (patch, "PATCH"),
            (delete, "DELETE"),
            (connect, "CONNECT"),
            (options, "OPTIONS"),
            (trace, "TRACE"),
        ):
            if operation_or_endpoint:
                operations[method] = (
                    operation_or_endpoint
                    if isinstance(operation_or_endpoint, Operation)
                    else Operation(operation_or_endpoint)
                )
        self.operations = operations
        super().__init__(  # type: ignore  # for Pylance
            path=path,
            # this needs to be an object so that Starlette
            # detects it as an ASGI app and passes us the raw Scope, Receive and Send
            # as well as not wrapping it in a threadpool
            endpoint=_PathApp(operations),  # type: ignore[arg-type]
            name=name or path,
            include_in_schema=include_in_schema,
            methods=list(operations.keys()),
        )



import sys
from typing import Any, NamedTuple

if sys.version_info < (3, 9):
    from typing_extensions import Annotated as Annotated  # noqa: F401
else:
    from typing import Annotated as Annotated  # noqa: F401


class Some(NamedTuple):
    value: Any



import json
from typing import Any, Dict, Optional

from starlette.responses import HTMLResponse

from xpresso.encoders import Encoder, JsonableEncoder


def get_swagger_ui_html(
    *,
    openapi_url: str,
    title: str,
    swagger_js_url: str = "https://cdn.jsdelivr.net/npm/swagger-ui-dist@4/swagger-ui-bundle.js",
    swagger_css_url: str = "https://cdn.jsdelivr.net/npm/swagger-ui-dist@4/swagger-ui.css",
    swagger_favicon_url: Optional[str] = None,
    oauth2_redirect_url: Optional[str] = None,
    init_oauth: Optional[Dict[str, Any]] = None,
    encoder: Encoder = JsonableEncoder(),
) -> HTMLResponse:
    if swagger_favicon_url:
        swagger_favicon_html = (
            f"""\n<link rel="shortcut icon" href="{swagger_favicon_url}">"""
        )
    else:
        swagger_favicon_html = ""

    html = f"""
    <!DOCTYPE html>
    <html>
    <head>
    <link type="text/css" rel="stylesheet" href="{swagger_css_url}">{swagger_favicon_html}
    <title>{title}</title>
    </head>
    <body>
    <div id="swagger-ui">
    </div>
    <script src="{swagger_js_url}"></script>
    <!-- `SwaggerUIBundle` is now available on the page -->
    <script>
    const ui = SwaggerUIBundle({{
        url: '{openapi_url}',
    """

    if oauth2_redirect_url:
        html += f"oauth2RedirectUrl: window.location.origin + '{oauth2_redirect_url}',"

    html += """
        dom_id: '#swagger-ui',
        presets: [
        SwaggerUIBundle.presets.apis,
        SwaggerUIBundle.SwaggerUIStandalonePreset
        ],
        layout: "BaseLayout",
        deepLinking: true,
        showExtensions: true,
        showCommonExtensions: true
    })"""

    if init_oauth:
        html += f"""
        ui.initOAuth({json.dumps(encoder(init_oauth))})
        """

    html += """
    </script>
    </body>
    </html>
    """
    return HTMLResponse(html)



from typing import Any, AsyncIterator, Awaitable, Dict, List, Union

from starlette.requests import HTTPConnection

import xpresso.openapi.models as openapi_models
from xpresso._utils.typing import Protocol


class SupportsExtractor(Protocol):
    def extract(
        self, connection: HTTPConnection
    ) -> Union[Awaitable[Any], AsyncIterator[Any]]:
        """Extract data from an incoming connection.

        The `connection` parameter will always be either a Request object or a WebSocket object,
        which are both subclasses of HTTPConnection.
        If you just need access to headers, query params, or any other metadata present in HTTPConnection
        then you can use the parameter directly.
        Otherwise, you can do `isinstance(connection, Request)` before accessing `Request.stream()` and such.

        The return value can be an awaitable or an async iterable (context manager like).
        The iterator versions will be wrapped with `@contextlib.{async}contextmanager`.
        """
        ...

    # __hash__ and __eq__ are required so that the dependency injection system can cache extracted values
    # (for example allowing the user to get multiple references to a request body without parsing twice)

    def __hash__(self) -> int:
        ...

    def __eq__(self, __o: object) -> bool:
        ...


Model = type
ModelNameMap = Dict[Model, str]


class SupportsOpenAPI(Protocol):
    def get_models(self) -> List[type]:
        """Collect all of the types that OpenAPI schemas will be
        produced from.

        Xpresso will then assign a schema name to each type and pass
        that back via the ModelNameMap parameter.

        This ensures that all schema models have a unique name,
        even if their Python class names conflict.
        """
        ...

    def modify_operation_schema(
        self,
        model_name_map: ModelNameMap,
        operation: openapi_models.Operation,
        components: openapi_models.Components,
    ) -> None:
        """Callback to modify the OpenAPI schema.

        Implementers should modify the operation and components as they see fit,
        but take care to not needlessly add keys or try to access keys which may not exist.

        When determining what string name to use to represent a model/schema (for example to add it to components/schemas)
        you MUST use the model_name_map parameter to find the name assigned for each type.
        For example:
        >>> components.schemas[model_name_map[MyModel]] = MyModel.get_schema()
        """
        ...



import inspect
from typing import Any, NamedTuple, Optional

from pydantic.error_wrappers import ErrorWrapper
from pydantic.fields import ModelField
from starlette.requests import HTTPConnection

from xpresso._utils.pydantic_utils import model_field_from_param
from xpresso.binders._binders.formencoded_parsing import Extractor as FormExtractor
from xpresso.binders._binders.formencoded_parsing import (
    InvalidSerialization,
    get_extractor,
)
from xpresso.binders._binders.pydantic_validators import validate_param_field
from xpresso.binders.api import SupportsExtractor
from xpresso.exceptions import RequestValidationError, WebSocketValidationError

ERRORS = {
    "websocket": WebSocketValidationError,
    "http": RequestValidationError,
}


class Extractor(NamedTuple):
    name: str
    field: ModelField
    extractor: FormExtractor

    def __hash__(self) -> int:
        return hash((self.__class__, self.name))

    def __eq__(self, __o: object) -> bool:
        return isinstance(__o, Extractor) and __o.name == self.name

    async def extract(
        self,
        connection: HTTPConnection,
    ) -> Any:
        try:
            extracted = self.extractor(
                name=self.name, params=connection.query_params.multi_items()
            )
        except InvalidSerialization:
            raise ERRORS[connection.scope["type"]](
                [
                    ErrorWrapper(
                        exc=TypeError("Data is not a valid URL encoded query"),
                        loc=tuple(("query", self.name)),
                    )
                ]
            )
        return validate_param_field(
            field=self.field,
            in_="query",
            name=self.name,
            connection=connection,
            values=extracted,
        )


class ExtractorMarker(NamedTuple):
    alias: Optional[str]
    explode: bool
    style: str

    def register_parameter(self, param: inspect.Parameter) -> SupportsExtractor:
        if self.style == "deepObject" and not self.explode:
            # no such thing in the spec
            raise ValueError("deepObject can only be used with explode=True")
        field = model_field_from_param(param)
        name = self.alias or param.name
        extractor = get_extractor(style=self.style, explode=self.explode, field=field)
        name = self.alias or field.alias
        return Extractor(field=field, name=name, extractor=extractor)



import typing

from di.api.providers import DependencyProvider
from di.concurrency import as_async
from di.dependent import Dependent
from di.dependent import Injectable as InjectableBase
from di.dependent import Marker

from xpresso._utils.typing import Literal

Scope = Literal["app", "connection", "endpoint"]
Scopes = (
    "app",
    "connection",
    "endpoint",
)


@typing.overload
def Depends(
    call: DependencyProvider,
    *,
    use_cache: bool = ...,
    wire: bool = ...,
    sync_to_thread: bool = ...,
    scope: Scope = ...,
) -> "BoundDependsMarker":
    ...


@typing.overload
def Depends(
    *,
    use_cache: bool = ...,
    wire: bool = ...,
    sync_to_thread: bool = ...,
    scope: Scope = ...,
) -> "DependsMarker[None]":
    ...


def Depends(
    call: typing.Any = None,
    *,
    use_cache: bool = True,
    wire: bool = True,
    sync_to_thread: bool = False,
    scope: typing.Optional[Scope] = None,
) -> typing.Any:
    if sync_to_thread:
        call = as_async(call)
    return DependsMarker(
        call=call,
        use_cache=use_cache,
        wire=wire,
        scope=scope,
    )


DependencyType = typing.TypeVar(
    "DependencyType", bound=typing.Optional[DependencyProvider]
)


class DependsMarker(Marker, typing.Generic[DependencyType]):
    def __init__(
        self,
        call: typing.Optional[DependencyType] = None,
        *,
        use_cache: bool = True,
        wire: bool = True,
        sync_to_thread: bool = False,
        scope: typing.Optional[Scope] = None,
    ) -> None:
        super().__init__(
            call=call,
            scope=scope,
            use_cache=use_cache,
            wire=wire,
        )
        self.sync_to_thread = sync_to_thread

    def as_dependent(self) -> Dependent[DependencyType]:
        call: "typing.Optional[DependencyProvider]"
        if self.sync_to_thread:
            if not self.call:
                raise ValueError(
                    "sync_to_thread can only be used if you explicitly declare the target function"
                )
            call = as_async(self.call)
        else:
            call = self.call
        return Dependent(
            call=call,  # type: ignore
            scope=self.scope,
            use_cache=self.use_cache,
            wire=self.wire,
            marker=self,
        )


BoundDependsMarker = DependsMarker[DependencyProvider]


class Injectable(InjectableBase):
    __slots__ = ()

    def __init_subclass__(
        cls,
        call: typing.Optional[DependencyProvider] = None,
        scope: typing.Optional[Scope] = None,
        use_cache: bool = True,
        **kwargs: typing.Any,
    ) -> None:
        return super().__init_subclass__(call, scope, use_cache, **kwargs)


class Singleton(InjectableBase):
    __slots__ = ()

    def __init_subclass__(
        cls,
        call: typing.Optional[DependencyProvider] = None,
        scope: Scope = "app",
        use_cache: bool = True,
        **kwargs: typing.Any,
    ) -> None:
        return super().__init_subclass__(call, scope, use_cache, **kwargs)



from typing import Any, Sequence

from di.api.dependencies import DependentBase
from di.api.scopes import Scope


def endpoint_scope_resolver(
    dep: DependentBase[Any],
    sub_dependent_scopes: Sequence[Scope],
    _: Sequence[Scope],
) -> Scope:
    """Resolve scopes by defaulting to "connection"
    unless a sub-dependency has an "endpoint" scope
    in which case we drop down to that scope
    """
    if dep.scope is not None:
        return dep.scope
    if "endpoint" in sub_dependent_scopes:
        return "endpoint"
    return "connection"


def lifespan_scope_resolver(
    dep: DependentBase[Any],
    sub_dependent_scopes: Sequence[Scope],
    _: Sequence[Scope],
) -> Scope:
    """Always default to the "app" scope"""
    if dep.scope is None:
        return "app"
    return dep.scope



import json
import typing

import xpresso.binders.dependents as dependents
import xpresso.openapi.models as openapi_models
from xpresso._utils.typing import Annotated, Literal
from xpresso.binders._binders import file_body, form_body, json_body, union

Example = typing.Union[openapi_models.Example, typing.Any]


def Json(
    *,
    examples: typing.Optional[typing.Dict[str, Example]] = None,
    description: typing.Optional[str] = None,
    decoder: json_body.SupportsJsonDecoder = json.loads,
    enforce_media_type: bool = True,
    consume: bool = True,
    include_in_schema: bool = True,
) -> dependents.BinderMarker:
    body_extractor_marker = json_body.ExtractorMarker(
        decoder=decoder,
        enforce_media_type=enforce_media_type,
        consume=consume,
    )
    body_openapi_marker = json_body.OpenAPIMarker(
        description=description,
        examples=examples,
        include_in_schema=include_in_schema,
    )
    return dependents.BinderMarker(
        extractor_marker=body_extractor_marker,
        openapi_marker=body_openapi_marker,
    )


def RawBody(
    *,
    media_type: typing.Optional[str] = None,
    enforce_media_type: bool = True,
    examples: typing.Optional[typing.Dict[str, Example]] = None,
    description: typing.Optional[str] = None,
    format: Literal["binary", "base64"] = "binary",
    consume: bool = True,
    include_in_schema: bool = True,
) -> dependents.BinderMarker:
    extractor_marker = file_body.ExtractorMarker(
        media_type=media_type,
        enforce_media_type=enforce_media_type,
        consume=consume,
    )
    openapi_marker = file_body.OpenAPIMarker(
        description=description,
        examples=examples,
        media_type=media_type,
        format=format,
        include_in_schema=include_in_schema,
    )
    return dependents.BinderMarker(
        extractor_marker=extractor_marker,
        openapi_marker=openapi_marker,
    )


def FormField(
    *,
    alias: typing.Optional[str] = None,
    style: openapi_models.FormDataStyles = "form",
    explode: bool = True,
) -> form_body.FormFieldMarker:
    extractor_marker = form_body.FormFieldExtractorMarker(
        alias=alias,
        style=style,
        explode=explode,
    )
    openapi_marker = form_body.FormFieldOpenAPIMarker(
        alias=alias,
        style=style,
        explode=explode,
    )
    return form_body.FormFieldMarker(
        extractor_marker=extractor_marker,
        openapi_marker=openapi_marker,
    )


def FormFile(
    *,
    alias: typing.Optional[str] = None,
    media_type: typing.Optional[str] = None,
    format: Literal["binary", "base64"] = "binary",
) -> form_body.FormFieldMarker:
    extractor_marker = form_body.FormFileExtractorMarker(
        alias=alias,
    )
    openapi_marker = form_body.FormFileOpenAPIMarker(
        alias=alias,
        format=format,
        media_type=media_type,
    )
    return form_body.FormFieldMarker(
        extractor_marker=extractor_marker,
        openapi_marker=openapi_marker,
    )


def Form(
    *,
    examples: typing.Optional[typing.Dict[str, Example]] = None,
    description: typing.Optional[str] = None,
    include_in_schema: bool = True,
) -> dependents.BinderMarker:
    extractor_marker = form_body.ExtractorMarker(
        media_type="application/x-www-form-urlencoded",
    )
    openapi_marker = form_body.OpenAPIMarker(
        description=description,
        examples=examples,
        media_type="application/x-www-form-urlencoded",
        include_in_schema=include_in_schema,
    )
    return dependents.BinderMarker(
        extractor_marker=extractor_marker,
        openapi_marker=openapi_marker,
    )


def Multipart(
    *,
    examples: typing.Optional[typing.Dict[str, Example]] = None,
    description: typing.Optional[str] = None,
    include_in_schema: bool = True,
) -> dependents.BinderMarker:
    extractor_marker = form_body.ExtractorMarker(media_type="multipart/form-data")
    openapi_marker = form_body.OpenAPIMarker(
        description=description,
        examples=examples,
        media_type="multipart/form-data",
        include_in_schema=include_in_schema,
    )
    return dependents.BinderMarker(
        extractor_marker=extractor_marker,
        openapi_marker=openapi_marker,
    )


def BodyUnion(
    *,
    description: typing.Optional[str] = None,
) -> dependents.BinderMarker:
    extractor_marker = union.ExtractorMarker()
    openapi_marker = union.BodyOpenAPIMarker(
        description=description,
    )
    return dependents.BinderMarker(
        extractor_marker=extractor_marker,
        openapi_marker=openapi_marker,
    )


if typing.TYPE_CHECKING:
    from pydantic import BaseModel

    from xpresso import UploadFile


# Convenience type aliases
_T = typing.TypeVar("_T")
_BaseModelT = typing.TypeVar("_BaseModelT", bound=typing.Union["BaseModel", None])
FromJson = Annotated[_BaseModelT, Json()]
_BinaryT = typing.TypeVar(
    "_BinaryT",
    bound=typing.Union[
        bytes,
        "UploadFile",
        typing.AsyncIterator[bytes],
        None,
    ],
)
FromRawBody = Annotated[_BinaryT, RawBody()]
FromFormField = Annotated[_T, FormField()]
_FormFileT = typing.TypeVar(
    "_FormFileT",
    bound=typing.Union[
        bytes, "UploadFile", typing.List[bytes], "typing.List[UploadFile]", None
    ],
)
FromFormFile = Annotated[_FormFileT, FormFile()]
FromFormData = Annotated[_BaseModelT, Form()]
FromMultipart = Annotated[_BaseModelT, Multipart()]
FromBodyUnion = Annotated[_T, BodyUnion()]



import typing
from typing import List, Type

from pydantic import BaseModel
from pydantic import ValidationError as PydanticValidationError
from pydantic import create_model
from pydantic.error_wrappers import ErrorWrapper
from starlette.exceptions import HTTPException as HTTPException  # noqa: F401
from starlette.status import HTTP_422_UNPROCESSABLE_ENTITY

_RequestErrorModel: Type[BaseModel] = create_model("Request")  # type: ignore
_WebSocketErrorModel: Type[BaseModel] = create_model("WebSocket")  # type: ignore


class RequestValidationError(PydanticValidationError):
    raw_errors: List[ErrorWrapper]

    def __init__(
        self,
        errors: typing.Sequence[ErrorWrapper],
        status_code: int = HTTP_422_UNPROCESSABLE_ENTITY,
    ) -> None:
        super().__init__(errors, _RequestErrorModel)
        self.status_code = status_code


class WebSocketValidationError(PydanticValidationError):
    def __init__(
        self,
        errors: typing.Sequence[ErrorWrapper],
    ) -> None:
        super().__init__(errors, _WebSocketErrorModel)
